<?php

namespace Drupal\connectid;

use ConnectID\Api\DataModel\CouponType;
use ConnectID\Api\DataModel\CouponTypeList;
use ConnectID\Api\DataModel\Order;
use ConnectID\Api\DataModel\OrderStatus;
use ConnectID\Api\DataModel\ProductTypeList;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\State\StateInterface;
use Drupal\Core\Url;
use League\OAuth2\Client\Token\AccessTokenInterface;
use Ramsalt\OAuth2\Client\Provider\ConnectId;
use Ramsalt\OAuth2\Client\Provider\Exception\InvalidApiResponseException;

/**
 * Class ApiService.
 */
class ApiService implements ApiServiceInterface {

  /**
   * @var \Drupal\Core\State\StateInterface
   */
  protected $stateService;

  /**
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /** @var \League\OAuth2\Client\Token\AccessTokenInterface */
  protected $clientAccessToken;

  /**
   * @var \ConnectID\Api\DataModel\ProductTypeList
   */
  protected $productsCache;

  /**
   * @var \ConnectID\Api\DataModel\CouponTypeList[]
   */
  protected $couponsCache;

  /**
   * The current active OAuth configuration.
   *
   * @var array
   */
  protected $oauthConfig;


  public function __construct(LoggerChannelInterface $logger, StateInterface $stateService) {
    $this->logger = $logger;
    $this->stateService = $stateService;
  }


  public function getProducts(): ProductTypeList {
    if (!$this->productsCache) {
      $this->productsCache = $this->getOauthClient()->getClientApiProducts();
    }

    return $this->productsCache;
  }


  public function getProductCoupons(string $productName): CouponTypeList {
    if (!isset($this->couponsCache[$productName])) {
      try {
        $this->couponsCache[$productName] = $this->getOauthClient()
          ->clientApi_getProductCoupons($productName);
      }
      catch (InvalidApiResponseException $exception) {
        watchdog_exception('connectid', $exception);
      }
    }

    return $this->couponsCache[$productName] ?? new CouponTypeList();
  }


  public function getCoupon(string $couponId): ?CouponType {
    [$productCode, $couponNumber] = explode(':', $couponId, 2);
    if (empty($productCode) || empty($couponNumber)) {
      throw new \InvalidArgumentException('Invalid coupon id: ' . $couponId);
    }
    $couponList = $this->getProductCoupons($productCode);
    return $couponList->offsetGet($couponId);
  }


  public function registerOrder(Order $order): Order {
    return $this->getOauthClient()->clientApi_registerOrder($order);
  }


  public function getOrderFulfillmentUrl(
    Order $order,
    Url $successUrl = NULL,
    Url $errorUrl = NULL
  ): Url {
    $successReturnUri = $errorReturnUri = \Drupal::request()->getUri();

    if ($successUrl) {
      $successReturnUri = $successUrl->setAbsolute()
        ->toString();
    }

    if ($errorUrl) {
      $errorReturnUri = $errorUrl->setAbsolute()
        ->toString();
    }

    $url_string = $this->getOauthClient()->getOrderFulfillmentUrl(
      $order->getOrderId(),
      $successReturnUri,
      $errorReturnUri
    );
    return Url::fromUri($url_string);
  }


  public function getOrderStatus(string $orderId): OrderStatus {
    return $this->getOauthClient()->clientApi_getOrderStatus($orderId);
  }

  protected function oauthConfig(): array {
    if (!$this->oauthConfig) {
      $moduleConfig = \Drupal::config('connectid.config');
      $this->oauthConfig = [
        'clientId'     => $moduleConfig->get('client_id'),
        'clientSecret' => $moduleConfig->get('client_secret'),
        'testing'      => $moduleConfig->get('testing_mode'),
      ];

      // Load overrides from the Key module.
      if (\Drupal::moduleHandler()->moduleExists('key')) {
        /** @var \Drupal\key\KeyRepositoryInterface $keyRepo */
        $keyRepo = \Drupal::service('key.repository');
        $keyName = $moduleConfig->get('client_secret');
        $this->oauthConfig['clientSecret'] = $keyRepo->getKey($keyName)
          ->getKeyValue();
      }
    }

    return $this->oauthConfig;
  }

  protected function getOauthClient(): ConnectId {
    return new ConnectId($this->oauthConfig());
  }

  protected function getClientToken(): AccessTokenInterface {
    // Static caching check first.
    if ($this->clientAccessToken === NULL) {
      // Try to load the token from the site State.
       $this->clientAccessToken = $this->stateService->get('connectid_client_token');
    }

    // Request a new token if we have an expired token or if we don't have one.
    if ($this->clientAccessToken === NULL || $this->clientAccessToken->hasExpired()) {
      $this->clientAccessToken = $this->getOauthClient()->getClientCredentialsAccessToken();
      $this->stateService->set('connectid_client_token', $this->clientAccessToken);
    }

    return $this->clientAccessToken;
  }

}
