<?php

namespace Drupal\connectid\Form;

use ConnectID\Api\DataModel\Address;
use ConnectID\Api\DataModel\CouponType;
use ConnectID\Api\DataModel\Order;
use ConnectID\Api\DataModel\OrderLine;
use DateTimeImmutable;
use Drupal\connectid\ApiServiceInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Routing\TrustedRedirectResponse;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a ConnectID API Access form.
 */
abstract class BaseOrderForm extends FormBase {

  /**
   * @var \Drupal\connectid\ApiServiceInterface
   */
  protected $api;


  public function __construct(ApiServiceInterface $apiService) {
    $this->api = $apiService;
  }


  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('connectid.api')
    );
  }


  /**
   * Submission handler for coupon selection.
   */
  public function ajaxCouponSelectionElement(
    array &$form,
    FormStateInterface $form_state
  ): array {
    return $form['product']['coupon_id'];
  }


  /**
   * Builds a Product + Coupon selection.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @return array
   *   Form API elements to select Product and Coupon.
   */
  protected function getProductSelectionElement(
    FormStateInterface $form_state
  ): array {
    $form['product'] = [
      '#type'  => 'fieldset',
      '#label' => $this->t('Product'),
    ];

    $product_options = [];
    $products = $this->api->getProducts();
    foreach ($products as $product) {
      $product_options[$product->getProduct()] = $product->getDescription();
    }
    $form['product']['product_id'] = [
      '#type'          => 'select',
      '#title'         => t('Product'),
      '#options'       => $product_options,
      '#default_value' => $form_state->getValue('product_id'),
      '#required'      => TRUE,
      '#ajax'          => [
        'callback'        => '::ajaxCouponSelectionElement',
        'disable-refocus' => TRUE,
        'event'           => 'change',
        'wrapper'         => 'coupon-wrapper',
        'progress'        => [
          'type'    => 'throbber',
          'message' => $this->t('Loading coupons...'),
        ],
      ],
    ];

    $form['product']['coupon_id'] = [
      '#type'          => 'select',
      '#title'         => t('Coupon'),
      '#options'       => [],
      '#default_value' => $form_state->getValue('coupon_id'),
      '#prefix'        => '<div id="coupon-wrapper">',
      '#suffix'        => '</div>',
      '#required'      => TRUE,
    ];

    // Prepare our textfield. check if the example select field has a selected option.
    if ($productId = $form_state->getValue('product_id')) {
      $coupons = $this->api->getProductCoupons($productId);
      $coupon_options = [];
      foreach ($coupons as $coupon) {
        $coupon_options[$coupon->getId()] = $coupon->getDescription();
      }

      $form['product']['coupon_id']['#options'] = $coupon_options;
    }
    else {
      $form['product']['coupon_id']['#empty_option'] = $this->t(
        '- Select a product to enable-'
      );
      $form['product']['coupon_id']['#disabled'] = TRUE;
    }


    return $form['product'];
  }


  /**
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   * @param string                               $title
   *   Fieldset title
   *
   * @return array
   */
  protected function getAddressFieldset(
    FormStateInterface $form_state,
    string $title = 'Address'
  ): array {
    $fieldset = [
      '#type'  => 'fieldset',
      '#title' => t($title),
      '#tree'  => TRUE,
    ];

    $fieldset['recipient_type'] = [
      '#type'    => 'radios',
      '#title'   => t('Recipient type'),
      '#options' => [
        'company' => t('Company'),
        'private' => t('Private'),
      ],
    ];

    $fieldset['companyName'] = [
      '#type'        => 'textfield',
      '#title'       => t('Company name'),
      '#placeholder' => t('Company name'),
    ];

    $fieldset['firstName'] = [
      '#type'        => 'textfield',
      '#title'       => t('First name'),
      '#placeholder' => t('First name'),
    ];

    $fieldset['lastName'] = [
      '#type'        => 'textfield',
      '#title'       => t('Last name'),
      '#placeholder' => t('Last name'),
    ];

    $fieldset['street'] = [
      '#type'        => 'textfield',
      '#title'       => t('Address'),
      '#placeholder' => t('Address'),
    ];

    $fieldset['postalCode'] = [
      '#type'        => 'textfield',
      '#title'       => t('Postal/ZIP code'),
      '#placeholder' => t('Postal/ZIP code'),
    ];

    $fieldset['postalPlace'] = [
      '#type'        => 'textfield',
      '#title'       => t('City/Area'),
      '#placeholder' => t('City/Area'),
    ];

    $fieldset['phone'] = [
      '#type'        => 'textfield',
      '#title'       => t('Telephone number'),
      '#placeholder' => t('Telephone number'),
    ];

    $fieldset['email'] = [
      '#type'        => 'textfield',
      '#title'       => t('Email address'),
      '#placeholder' => t('Email address'),
    ];

    return $fieldset;
  }


  protected function getPaymentMethodElement(
    FormStateInterface $form_state
  ): array {
    // TODO: Make this configurable
    $payment_methods = [
      Order::PAYMENT_TYPE_CARD    => t('Credit Card'),
      Order::PAYMENT_TYPE_INVOICE => t('Invoice'),
      Order::PAYMENT_TYPE_SMS     => t('SMS'),
    ];

    return [
      '#type'     => 'radios',
      '#title'    => t('Payment method'),
      '#options'  => $payment_methods,
      '#required' => TRUE,
    ];
  }


  protected function createOrderFromSubmission(FormStateInterface $form_state
  ): Order {
    $order = new Order();

    $requestTime = DateTimeImmutable::createFromFormat(
      'U',
      \Drupal::time()->getRequestTime()
    );
    $order->withOrderDate($requestTime)
      ->withPaymentMethod($form_state->getValue('paymentMethod'))
      ->withCurrency('NOK');

    $billing_address = $this->createConnectIdAddressFromInput(
      $form_state->getValue('billing_address')
    );
    $order->withPayer($billing_address);

    [$couponCode, $couponNum] = explode(
      ':',
      $form_state->getValue('coupon_id')
    );
    $coupon = new CouponType();
    $coupon->withCouponCode($couponCode)
      ->withCouponNumber((int) $couponNum)
      ->withCouponPrice(123.0);

    $order->withOrderAmount($coupon->getCouponPrice());

    $order_line_alpha = new OrderLine();
    $order_line_alpha->withQuantity(1)
      ->withProductSpecType('coupon')
      ->withProductSpecCode($coupon->getCouponCode())
      ->withProductSpecNo($coupon->getCouponNumber())
      ->withUnitPrice($coupon->getCouponPrice());

    return $order->withOrderLine($order_line_alpha);
  }


  /**
   * Creates an Address object from an array of submitted data.
   *
   * The data in the array is expected to reflect a fieldset as generated
   *  by BaseOrderForm::getAddressFieldset()
   *
   * @see self::getAddressFieldset()
   *
   * @param array $form_values
   *   The array of data.
   *
   * @return \ConnectID\Api\DataModel\Address
   *   The Address object ready to use.
   */
  protected function createConnectIdAddressFromInput(array $form_values): Address {
    if (!empty($form_values['email'])) {
      $form_values['emails'] = [
        $form_values['email'],
      ];
      unset($form_values['email']);
    }

    if (!empty($form_values['phone'])) {
      $form_values['phoneNumbers'] = [
        $form_values['phone'],
      ];
      unset($form_values['phone']);
    }

    unset($form_values['recipient_type']);

    return Address::create($form_values);
  }


  /**
   * Submits an order to the API.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   * @param \ConnectID\Api\DataModel\Order       $order
   *
   * @return \ConnectID\Api\DataModel\Order
   *   The submitted order which includes an Order identifier.
   */
  protected function submitOrder(
    FormStateInterface $form_state,
    Order $order
  ): Order {
    $order = $this->api->registerOrder($order);

    if ($order->getOrderId() === NULL) {
      throw new \RuntimeException(
        'Order not submitted. Error handling not implemented'
      );
    }
    return $order;
  }


  /**
   * This sets the form redirect to the final "order fulfillment" page.
   *
   * This page is hosted on ConnectID and therefore requires us to create a
   *  TrustedRedirectResponse ad-hoc to allow the redirect.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   * @param \ConnectID\Api\DataModel\Order       $order
   */
  protected function setOrderFulfillRedirect(
    FormStateInterface $form_state,
    Order $order
  ): void {
    $form_state->setResponse(
      new TrustedRedirectResponse(
        $this->api->getOrderFulfillmentUrl($order)->getUri()
      )
    );
  }

}
